/* Draw a heightfield on the screen using X-Windows (X11 graphics) */
/* JPB 9/27/95 */
/* mods JPB 10/26/95, 1/16/96 */

#include <stdio.h>                      /* fopen() sprintf() etc. */
#include <stdlib.h>                     /* strtod() */
#include <math.h>                       /* math stuff */
#include <string.h>                     /* strcpy() */
#include <ctype.h>                      /* tolower() */
#include "hl.h"
#include "hf_io.h"
#define MAIN
#include "typy.h"

#define ESC '/033'

#define FALSE 0
#define TRUE 1
#define D double

#define Elcol(vv, xq, yq)  vv[((yq) * xsize2) + (xq)]

/* -------- X-Window graphics routines ----------- */
VOID Clear();
VOID X_Point(INT x,INT y);   /* drawable point */
VOID X_Dot(INT x,INT y, INT Color);   /* colored drawable pixel */
VOID X_Poly(INT N,POINT2 point[]);    /* unfilled polygon */
VOID X_FillPoly(INT N,POINT2 point[],INT CO,INT CV);
/* ------------------------------------------------- */


		/* draw two triangles joined along hypotenuse */
void tri3d(D x0, D z0,D x1,D z1,D y0,D y1,D y2,D y3, int shader, int col);
void drawhf(PTYPE *d1, int maxx, int maxz, int s, int t);
int graph_setup(int mode, int *xsize, int *ysize);
void graph_clear();   /* clear graphics screen */
void pr_ghelp();              /* print help message */
void pr_vars(char *fname, int xsize, int ysize, int xgsize, int ygsize);
int getline(char s[], int lim);       /* read in one line */

int show_grid(hfield hf, char insp[MAXTOK][TOKLEN], int tokens);
int show_flat(hfield hf, char insp[MAXTOK][TOKLEN], int tokens);
int hf_show1(hfield hf,int xw0,int yw0, int xw1, int yw1);

extern hfield *hst;      /* stack of heightfields */
extern int hp;           /* pointer to top of stack: hst[hp] */
extern int stacksize;    /* size of HF stack */
extern int nskip;        /* number of hf elements to skip */
extern int v_tile;       /* view tiled hf? */
extern int xgsize,ygsize; /* size of graphics window */
extern int graph_mode;   /* initial graphics resolution mode (DOS) */
extern int noxinit;      /* if X window has been created */
extern int shader;       /* 0 wireframe 1 height 2 slope 3 lightsource */
extern int hist_on;      /* if history being recorded */
extern D sl_min;         /* minimum slope (shading offset) */
extern D sl_scale;       /* slope scale factor (shading scale) */
extern PT3 light;  /* position of light source */
extern PT3 eye;    /* position of viewer */

static int vfirst = TRUE;       /* if this is first call to show_grid() */

D zcenter; 
D zfar, znear;
D zsf;      /* z coordinate scale factor */
D hs;      /* vertical (y) coordinate scale factor */
D ycorr;  /* y offset correction */
D zeps;   /* z epsilon */
D pinv2 = 0.15915;
int yoff;  /* screen coords y-offset */
short sxmid;  /* center coordinates of view window (eg 640x480) */
short symid;
int maxcol = 127;  /* maximum color index */
D bright;           /* intensity of light  */
int maxcon;         /* maximize contrast? (boolean) */
int cmin, cmax;     /* max and min color index value in this image */
int erasewin = TRUE;       /* erase window between displays? */
int updatewin = TRUE;
int icolormap = FALSE;     /* index color map sucessfully read in */
PTYPE *d2;
U xsize2, ysize2; /* colormap array size */

  /* 3D projection of heightfield (triangulation) */

int show_grid(hfield hf,           /* heightfield to display */
    char insp[MAXTOK][TOKLEN],     /* tokens array, each TOKLEN-1 long max. */
	      int tokens           /* number of tokens present */
  )
{
  PTYPE *d1;          /* HF array */
  U xsize, ysize;   /* HF dimensions */
  int tt;
  int t_dur = 0;
  char c;
  D eyinc;
  D lmag;         /* magnitude of a vector */
  char instr[120];  /* input string */
  char fname[80];   /* input file name */
  char cstr[80];
  int draw = TRUE;

  if( ( d1 = hf.a) == (PTYPE *)NULL) {
    printf("No matrix to view.\n");
    return(1);
  }
  xsize = hst[hp].xsize;
  ysize = hst[hp].ysize;

  tt = 1;
  if (vfirst) {
     nskip = xsize/25;
     if (nskip < 1) nskip = 1;
     vfirst = FALSE;
  }
  if (tokens > 1) {
   do {
    if (tolower(insp[tt][0])=='p') sscanf(&insp[tt][1],"%d",&nskip);
    else if (tolower(insp[tt][0])=='w') sscanf(&insp[tt][1],"%d",&t_dur);
    else if (tolower(insp[tt][0])=='s') sscanf(&insp[tt][1],"%d",&shader);
    else if (tolower(insp[tt][0])=='t') sscanf(&insp[tt][1],"%d",&v_tile);
    else if (tolower(insp[tt][0])=='x') sscanf(&insp[tt][1],"%d",&xgsize);
    else if (tolower(insp[tt][0])=='y') sscanf(&insp[tt][1],"%d",&ygsize);
    else if (tolower(insp[tt][0])=='g') 
      sscanf(&insp[tt][1],"%d",&graph_mode);
    else {
      printf("view: %s not recognized.\n",insp[tt]);
    }
   } while (++tt < tokens);
  }

  if (t_dur<0) t_dur=0;        /* sanity checks; force correct range */
  if (nskip<1) nskip=1;
  if (shader > 3) shader = 3;
  if (shader < 0) shader = 0;
  if (graph_mode>3) graph_mode = 3;
  
		/* set up graphics mode */
 if (noxinit) { 
    if (init_graph(xgsize,ygsize,maxcol+1)!=0) return(1);
    noxinit = FALSE;
  } else {
     MaxX = GetMaxX(); MaxY = GetMaxY();
     xgsize = MaxX;    ygsize = MaxY;
  }

  printf("%s is a (%d x %d) heightfield.\n",hf.name,xsize,ysize);
  if (nskip != 1) {
      printf("Note: For faster preview, P%d has been set.\n",nskip);
      printf("To see full resolution, enter P1.\n");
    }

  MinX = 0; MinY = 0;
  MaxX = GetMaxX(); 
  MaxY = GetMaxY();
  xgsize = MaxX;
  ygsize = MaxY;
  sxmid = xgsize / 2;  /* middle of screen in screen coords */
  symid = ygsize / 2;
  zsf = 0.3;      /* z coordinate scale factor */
  hs = 0.3;      /* vertical (y) coordinate scale factor */
  zeps = 0.001;


	/* viewpoint 0,0,0 in +z direction */

  eyinc = 0.8;
		     /* light direction vector */
  lmag = sqrt(light.x*light.x + light.y*light.y + light.z*light.z);
  light.x /= lmag;
  light.y /= lmag;
  light.z /= lmag;

  bright = 0.8;  /* lightsource intensity */

  draw = TRUE;
  maxcon = FALSE;
  cmin = 0; cmax = maxcol;
  printf("Type 'h' for help.");
  if (t_dur>0) printf("  %d second pause...",t_dur);
  printf("\n");
  /* ------------------------------------main interp. loop ------- */
  do {
    zcenter = -eye.z;
    ycorr = 2*eye.y/(zcenter);
    xsize = hst[hp].xsize;
    ysize = hst[hp].ysize;
    if (draw && updatewin) {
      MaxX = GetMaxX();      MaxY = GetMaxY();
      xgsize = MaxX;         ygsize = MaxY;
      sxmid = xgsize / 2;  /* middle of screen in screen coords */
      symid = ygsize / 2;
      yoff = (int) (symid*(ycorr + eye.y/zcenter)); 
      if (erasewin) Clear();
      drawhf(d1,xsize,ysize,shader,v_tile);
      if (t_dur>0) {
	sleep((U)t_dur);
	return(0);
      }
    }

    draw = TRUE;
    do {
      hist_on = FALSE;
      printf("view>");            /* input prompt for user entry */
      getline(instr,80);
      hist_on = TRUE;
    } while (instr[0] == '\0');
    c = tolower(instr[0]);      /* first character */
    sscanf(instr,"%s",cstr);  /* first command word */    

    if (c == '+') hs *= 1.3;
    else if (c == '-') hs /= 1.3;
    else if (c == 'u') eye.y -= eyinc;
    else if (c == 'd') eye.y += eyinc;
    else if (c == 'b') eye.z -= 1.0;

    else if (c == '>') bright *= 1.2;
    else if (c == '<') bright /= 1.2;
    else if (keymatch(instr,"sleep",3)) {
	sleep(5);
    }
    else if (strcmp(cstr,"map")==0) {     /* read in elevation index map */
      sscanf(instr,"%s %s",cstr,fname);
      printf("This feature not yet supported; sorry.\n");
/*      icolormap = TRUE;
      shader = 4;  */       /* use new image map shader type */
    }
    else if (c == 'm') maxcon = !maxcon;
    else if (c == 'w') updatewin = !updatewin;
    else if (c == 'v') {
       pr_vars(hst[hp].name, xsize, ysize, xgsize, ygsize);
       draw = FALSE;
     }
    else if (c == 'l') {
      maxcon = FALSE;      /* reset contrast limits */
      sscanf(&instr[1],"%lf %lf %lf",&light.x,&light.y,&light.z);
      lmag = sqrt(light.x*light.x + light.y*light.y + light.z*light.z);
      light.x /= lmag;
      light.y /= lmag;
      light.z /= lmag;
    }
    else if (c == 'e') {
      maxcon = FALSE;      /* reset eye position */
      sscanf(&instr[1],"%lf %lf %lf",&eye.x,&eye.y,&eye.z);
      if (eye.z > -1.01) eye.z = -1.01;
    }
    else if (c == 'f') {
      eye.z += 0.5;
      if (eye.z >= -1.1) eye.z = -1.1;
    }
    else if ((c == 'h') || (c == '?')) {
      pr_ghelp();
      draw = FALSE;
    }
    else if (c == 't') v_tile = !(v_tile);
    else if (c == 's') {
     maxcon = FALSE;
     shader = (int)atol(&instr[1]);
     if (shader > 5) shader = 5;
     if (shader < 0) shader = 0;
     if ((shader > 3) && (!icolormap)) {
       printf("Must load colormap first ( map <filename> ).\n");
       shader = 0;
     }
    }
    else if (c == 'r') {
/*     maxcon = FALSE;
     sscanf(&instr[1],"%s",fname);
     if (readn_hf(&d1, &xsize, &ysize, fname) != 0) {
	 printf("Error reading %s\n",fname);
	 strcpy(fname,"previous image");
     }
     printf("%s: (%d x %d)\n",fname,xsize,ysize); */
     printf("sorry, this feature not supported yet.\n");
    }
    else if (c == 'p') {
     nskip = (int)atol(&instr[1]);
     if (nskip < 1) nskip = 1;
    }
    else if (c == 'q') { /* quitting */ }
    else {
      printf("%s not understood.\n",instr);
      draw = FALSE;
    }
  } while (c != 'q');

/*    close_graph();  */        /* close window, de-allocate memory */  
  return(0);
}

void pr_vars(char *fname, int xsize, int ysize, int xgsize, int ygsize)
	      /* print variable values */
{
   printf("eye coords:   %f %f %f\n",eye.x,eye.y,eye.z);
   printf("light vector: %f %f %f\n",light.x,light.y,light.z);
   printf("vertical scale: %f\n",hs);
   printf("lightsource I:  %f\n",bright);
   printf("contrast maximization is %s.\n",maxcon ? "ON" : "OFF");
   printf("color min,max: %d - %d\n",cmax,cmin);
   printf("grid spacing P%d",nskip);
   printf(" update is %s",updatewin ? "ON":"OFF");
   printf(" clear is %s.\n",erasewin ? "ON":"OFF");
   printf("%s is (%d x %d)  window: %d x %d\n",fname,xsize,ysize,
	  xgsize,ygsize);
}


void pr_ghelp()              /* print help message */
{
    printf("----------- Xgrid Help ------------------------\n");
    printf("---  There are six shading options as follows:\n");
    printf("S0  wireframe   S1  height-shading\n");
    printf("S2  slope       S3  lightsource\n");
    printf("S4, S5  external elevation index map\n");
    printf("L x y z   vector to lightsource (affects S3 mode only)\n");
    printf("E x y z   position of viewer (eye)\n");
    printf("B/F   move eye position back/forward.\n");
    printf("U/D   move eye position up/down\n");
    printf("M     toggle 'maximize contrast' mode\n");
    printf("\n");
    printf(">/<   brighter/darker (S3 mode only)\n");
    printf("+/-   change vertical scaling of heightfield.\n");
    printf("T     tile heightfield in a 2x3 array\n");
    printf("Pn    draw only every nth line, for faster preview.\n");
    printf("V     print variables   W update?    H help\n");
    printf("R fn  read in a new file, eg   R output.pgm\n");
    printf("Map fn read in an elevation colormap for S4 mode\n");
    printf("Q     quit\n");
    printf("\n");
}

/* =============================================================== */
/* project two triangles, given 3D coordinates, on a 2D viewscreen */
/* uses gobal vars:
 *   hs : vertical heightfield scaling factor
 *   ex,ey,ez : location of eye (viewpoint)
 *   symid : y coordinate midpoint
 *   ycorr : y correction factor
 *   yoff  : y offset in screen units 
 *   zcenter : center to 'rotate' appparent hf around
 *   maxcol  : maximum color value 
 *   bright  : lightsource intensity 
 * ===============================================================
 */

/*  shader: 0 = wireframe, 1 = height color, 2 = slope, 3 = lightsource */
/*          4 = extern. map */

void tri3d(D x0, D z0,D x1,D z1,D y0,D y1,D y2,D y3, int shader, int col)
{
 D zmod0, zmod1; 
 D slope1;
 D slope2; 
 int c1, c2, c3;  /* colors */
 D nx,ny,nz,nmag;   /* normal vector */
 D ctheta;          /* cos(theta)  theta = angle between light and normal */

 POINT2 p[4];       /* three sets of points in 2D -- but must alloc 4 for edge lines */
 c3 = c2 = c1 = 0;            /* make compiler happy */ 
 /*
 printf("t3d %1.3f %1.3f %1.3f %1.3f",x0,z0,x1,z1);
 printf(" %1.3f %1.3f %1.3f %1.3f ",y0,y1,y2,y3);
 printf(" * %d %d ",shader,col); 
 fflush(stdout);
 */

 if (shader == 1) {  /* height, before transform */
   c1 = maxcol*(y0 + y1 + y2 + y3)/4;
   c2 = c1;
 } 

 if ((shader == 4) || shader == 5) {
    c1 = col;
    c2 = col;
 }

 /* x0,y0,z0 should be in the range -1..1 */
 /* transform to eye coordinate (eye.x,eye.y,eye.z) */
 
 y0 *= -hs; y1 *= -hs; y2 *= -hs; y3 *= -hs;                          
 x0 -= eye.x; y0 -= eye.y; z0 -= eye.z;
 x1 -= eye.x; y1 -= eye.y; z1 -= eye.z;
 y2 -= eye.y; y3 -= eye.y; 
 if ((z1 <= zeps) || (z0 <= zeps)) return;  /* foreground clipping */

 if (shader == 2) {
   slope1 = ATAN2( (z0-z1)*(x1-x0),(z1-z0)*(y2-y0) );
   slope2 = ATAN2( (z0-z1)*(x1-x0),(z1-z0)*(y1-y3) ); 
   c1 = (int) maxcol*(slope1-sl_min)*sl_scale;
   c2 = (int) maxcol*(slope2-sl_min)*sl_scale;
   
 } else if (shader == 3) { /* compute normal, dot product with lightsource */

   nx = (z0-z1)*(y2-y0);
   ny = (z1-z0)*(x1-x0);
   nz = (x1-x0)*(y2-y1);
   nmag = sqrt(nx*nx + ny*ny + nz*nz);
   /* printf("nxyz %1.3f %1.3f %1.3f  nmag %1.3f ",nx,ny,nz,nmag); fflush(stdout); */
   if (nmag==0) { 
       ctheta = 0;
       printf("nmag=zero ");
       fflush(stdout);
   }
     else ctheta = (nx*light.x + ny*light.y + nz*light.z) / nmag;
   
   c1 = (int) (maxcol>>1)*(bright*(ctheta + 1.0));
   
   nx = -(z0-z1)*(y3-y1);
   /* ny = (z1-z0)*(x1-x0); */   /* duplicates line above */
   nz = (x0-x1)*(y3-y0);
   nmag = sqrt(nx*nx + ny*ny + nz*nz);
   if (nmag==0) ctheta = 0;
    else   ctheta = (nx*light.x + ny*light.y + nz*light.z) / nmag;
   c2 = (int) (maxcol>>1)*(bright*(ctheta + 1.0));

 } else if (shader == 0) {  /* shader = 0: wireframe */
   c1 = maxcol;  /* white, we hope. brightest color, anyway */
 }

 if (maxcon &&(cmax!=cmin)) {          /* if maximize_contrast is TRUE... */
   c1 = (int) (((D)maxcol/(cmax-cmin))*(c1-cmin));
   c2 = (int) (((D)maxcol/(cmax-cmin))*(c2-cmin));
 } 
 if (c1 < 0) c1 = 0;
 if (c1 > maxcol) c1 = maxcol;
 if (c2 < 0) c2 = 0;
 if (c2 > maxcol) c2 = maxcol;

 if (!maxcon) {
   if (c1 > cmax) cmax = c1;
   if (c1 < cmin) cmin = c1;
 }

 zmod0 = z0 * zsf;   /* z1 multiplied by z scale factor */
 zmod1 = z1 * zsf;
/*
 printf("zmod: %f %f\n",zmod0,zmod1);
  getch(); */
  /* sxmid, symid are midpoints of screen, eg sxmid=320 for 640x480 scn. */

 p[0].x = (int) (sxmid*x0/zmod0) + sxmid;  /* x,y,z 0 */
 p[0].y = (int) (symid*y0/zmod0) + symid + yoff;
 
 p[1].x = (int) (sxmid*x1/zmod1) + sxmid;  /* x,y,z 1 */
 p[1].y = (int) (symid*y1/zmod1) + symid + yoff;

 p[2].x = (int) (sxmid*x1/zmod0) + sxmid; /* x,y,z 2 */
 p[2].y = (int) (symid*y2/zmod0) + symid + yoff;
 
 if( ((p[0].x==p[1].x) && (p[0].y==p[1].y)) ||
     ((p[0].x==p[2].x) && (p[0].y==p[2].y)) ||
     ((p[1].x==p[2].x) && (p[1].y==p[2].y))  )  {
       X_Point(p[0].x,p[0].y);
   } else {
     if (shader == 0) {
       X_FillPoly(3,p,maxcol,0);
     }
     else if (shader == 1 || shader == 4) {
       c3 = c1 / 2;
       if (c1 < (maxcol/2)) c3 += maxcol/2;
       X_FillPoly(3,p,c3,c1);
     } else 
       X_FillPoly(3,p,c1,c1);   
   }

 p[1].x = (int) (sxmid*x0/zmod0) + sxmid;  /* x,y,z 0 */
 p[1].y = (int) (symid*y0/zmod0) + symid + yoff;

 p[0].x = (int) (sxmid*x1/zmod1) + sxmid;  /* x,y,z 1 */
 p[0].y = (int) (symid*y1/zmod1) + symid + yoff;

 p[2].x = (int) (sxmid*x0/zmod1) + sxmid; /* x,y,z 3 */
 p[2].y = (int) (symid*y3/zmod1) + symid + yoff; 

if( ((p[0].x==p[1].x) && (p[0].y==p[1].y)) ||
     ((p[0].x==p[2].x) && (p[0].y==p[2].y)) ||
     ((p[1].x==p[2].x) && (p[1].y==p[2].y))  )  {
       X_Point(p[0].x,p[0].y);
   } else {
     if (shader == 0)  {
       X_FillPoly(3,p,maxcol,0);
     }
     else if (shader == 1 || shader == 4) {
       c3 = c1 / 2;
       if (c1 < (maxcol/2)) c3 += maxcol/2;
       X_FillPoly(3,p,c3,c1);
     } else 
       X_FillPoly(3,p,c2,c2);
   }

} /* end tri3d() */

/* draw a heightfield, back to front, as a series of triangles */
/* uses global variable nskip as hf index increment. */

void drawhf(PTYPE *d1, int xsize, int ysize, int shader, int tile)
{
  int ix,iz,izi;
  int ymin, ymax, xmin, xmax;
  int noffx, noffz;
  int col = 0;
  D xd2, yd2;
  D x0, y0, z0, x1, y1, z1;
  D y2, y3;

  
  xd2 = xsize / 2.0;
  yd2 = ysize / 2.0;
  if (tile) {
    ymin = 0;
    ymax = (2*ysize) - 1;
    xmin = -xsize + 1;
    xmax = 2*xsize - 1;
  } else {
    ymin = 0;
    ymax = ysize-1;
    xmin = 0;
    xmax = xsize-1;
  }
  if (!maxcon) {
    cmax = 0; 
    cmin = maxcol;     /* initialize global lims to extrema */
  }
  /* printf("drawhf %d %d %d %d\n",xmin,xmax,ymin,ymax); */

		     /* loop over the two array indices */        
  /* z = into the screen  x = to right across the screen  y = vertically upwards */

    for ( iz=ymin;iz<ymax;iz += nskip) {
	  izi = (ymax - iz) + ymin;
	  if ((iz+nskip) < ymax) noffz = nskip;
	     else noffz = ymax-iz;
	  for ( ix=xmin;ix<xmax;ix+= nskip) {
	    if ((ix+nskip) < xmax) noffx = nskip;
	       else noffx = xmax-ix;
	    x0 = (ix-xd2)/xd2; z0 = (izi-yd2)/yd2; 
	    x1 = (ix+noffx-xd2)/xd2; z1 = (izi-noffz-yd2)/yd2; 
	    y0 = (D) Elmod(d1,ix,iz);
	    y1 = (D) Elmod(d1,ix+noffx,iz+noffz);
	    y2 = (D) Elmod(d1,ix+noffx,iz);
	    y3 = (D) Elmod(d1,ix,iz+noffz);
	    if (shader==4 || shader==5) {
	      col = maxcol * Elcol(d2,ix*xsize2/xsize,iz*ysize2/ysize);
	    }
	    tri3d(x0,z0,x1,z1,y0,y1,y2,y3,shader,col); 
	  } /* end for iz */
    } /* end for ix */
    Gflush();             /* flush events to X disp. */
} /* end drawhf() */

		/* show flat 2D image as greyscale */
int show_flat(hfield hf, char insp[MAXTOK][TOKLEN], int tokens)
{
 PTYPE *d1;
 int t_dur = 0;
 int gr_clear = TRUE;
 int show4 = FALSE;
 int tt;

 if( ( d1 = hf.a) == (PTYPE *)NULL) {
   printf("No matrix to show.\n");
   return(1);
 }

  tt = 1;
  if (tokens > 1) {
   do {
    if (tolower(insp[tt][0])=='p') sscanf(&insp[tt][1],"%d",&nskip);
    else if (keymatch(insp[tt],"nc",2)) gr_clear = FALSE;
    else if (tolower(insp[tt][0])=='4') show4 = TRUE;
    else if (tolower(insp[tt][0])=='w') sscanf(&insp[tt][1],"%d",&t_dur);
    else if (tolower(insp[tt][0])=='t') sscanf(&insp[tt][1],"%d",&v_tile);
    else if (tolower(insp[tt][0])=='x') sscanf(&insp[tt][1],"%d",&xgsize);
    else if (tolower(insp[tt][0])=='y') sscanf(&insp[tt][1],"%d",&ygsize);
    else if (tolower(insp[tt][0])=='g') 
      sscanf(&insp[tt][1],"%d",&graph_mode);
    else {
      printf("show: %s not recognized.\n",insp[tt]);
    }
   } while (++tt < tokens);
  }


		/* set up graphics mode */
 if (noxinit) { 
    if (init_graph(xgsize,ygsize,maxcol+1)!=0) return(1);
    noxinit = FALSE;
  } else {
     MaxX = GetMaxX(); MaxY = GetMaxY();
     xgsize = MaxX;    ygsize = MaxY;
     if (gr_clear) Clear();
  }

 if (show4) {   /* show four HFs at once */
    DECSTACK; DECSTACK; DECSTACK;
    hf_show1(hst[hp],xgsize/2,ygsize/2,xgsize,ygsize);
    INCSTACK;
    hf_show1(hst[hp],0,ygsize/2,xgsize/2,ygsize);
    INCSTACK;
    hf_show1(hst[hp],xgsize/2,0,xgsize,ygsize/2);
    INCSTACK;
    hf_show1(hst[hp],0,0,xgsize/2,ygsize/2);

 } else {       /* just show one HF */
    hf_show1(hst[hp],0,0,xgsize,ygsize);
 }

 /* printf("[%d x %d] %1.2f ... %1.2f\n",xsize,ysize,hf.min,hf.max); */

 if (t_dur>0) {
     sleep((U)t_dur);
     return(0);
 }
/* printf("show> ");
 getline(instr,80); */
 return(0);
}

/* hf_show1()  show a 2D representation of HF in greyscale in an X window */
/*      parameters: heightfield, xmin,ymin,xmax,ymax */
int hf_show1(hfield hf,int xw0,int yw0, int xw1, int yw1)
{
 PTYPE *d1;         /* ptr to HF data */
 int xsize,ysize;   /* HF size */
 int x,y,xx,yy;     /* indexing variables */
 int xpels, ypels;  /* # of pixels of image drawn */
 D off, scale;      /* HF value offset & scale to fit greyscale range */
 size_t memsize;    /* # bytes to allocate for image */
 char *data;        /* image data bytes */
 D imin, imax;      /* range of data in matrix */

 d1 = hf.a;
 if (d1==NULL) return(1);
 xsize = hf.xsize;
 ysize = hf.ysize;
 off = -hf.min;
 scale = (maxcol)/ (hf.max - hf.min);
 xpels = xw1-xw0;
 ypels = yw1-yw0;

 if (hf.c) {  /* ------------ matrix is complex --------------- */

  memsize = xpels*ypels*sizeof(char)/4;  /* real part of mx */

  data = (char *)malloc(memsize);  /* bytes to display */
  if (data == NULL) {
    printf("Could not allocate memory for display.\n");
    return(1);
  }
  memset(data,0,memsize);

   for (y=0;y<ypels/2;y++) {
    yy = 2*(y*ysize)/ypels;
    for (x=0;x<xpels/2;x++) {
     xx = 2*(x*xsize)/xpels;
     data[x + xpels*y/2] = (char)((El(d1,xx,yy) + off) * scale);
    }
   }
   X_PutImagemap(data, xpels/2, ypels/2, xw0,yw0+ypels/4); 

       /*-----------------  display complex part ----------- */

   data = (char *)malloc(memsize);  /* bytes to display */
   if (data == NULL) {
     printf("Could not allocate memory for display.\n");
     return(1);
   }
   memset(data,0,memsize);
   i_minmax(hf,&imin,&imax);
   off = -imin;
   scale = (maxcol)/ (imax - imin);

   for (y=0;y<ypels/2;y++) {
    yy = 2*(y*ysize)/ypels;
    for (x=0;x<xpels/2;x++) {
     xx = 2*(x*xsize)/xpels;
     data[x + xpels*y/2] = (char)((Im(d1,xx,yy) + off) * scale);
    }
   }
   X_PutImagemap(data, xpels/2, ypels/2, xw0+xpels/2,yw0+ypels/4); 
                             /* draw image in window */

 } else {
  memsize = xpels*ypels*sizeof(char);
  data = (char *)malloc(memsize);  /* bytes to display */
  if (data == NULL) {
    printf("Could not allocate memory for display.\n");
    return(1);
  }
  memset(data,0,memsize);

  /*  xoff = (xgsize-ygsize)/2;  yoff = 0; */

   for (y=0;y<ypels;y++) {
    yy = (y*ysize)/ypels;
    for (x=0;x<xpels;x++) {
     xx = (x*xsize)/xpels;
     data[x + xpels*y] = (char)((Elmod(d1,xx,yy) + off) * scale);
    }
   }

  X_PutImagemap(data, xpels, ypels, xw0,yw0); /* draw image in window */

 } /* end else */

 /* note that "data" is actually free()'d by the X_PutImagemap call. Scary. */

  return(0);  
} /* hf_show1() */




